import errno
import locale
import os
import stat
from contextlib import contextmanager
from io import BytesIO, StringIO
from os import strerror
from typing import IO, Iterator, List, NamedTuple, Optional, Tuple

from typing_extensions import Self

if hasattr(locale, "getencoding"):
    DEFAULT_ENCODING = locale.getencoding()
else:
    DEFAULT_ENCODING = locale.getpreferredencoding(False)


def file_not_found_error(*args):
    """return FileNotFoundError"""

    return OSError(errno.ENOENT, strerror(errno.ENOENT), *args)


def not_a_directory_error(*args):
    """return NotADirectoryError"""

    return OSError(errno.ENOTDIR, strerror(errno.ENOTDIR), *args)


def file_exists_error(*args):
    """return FileExistsError"""

    return OSError(errno.EEXIST, strerror(errno.EEXIST), *args)


def is_a_directory_error(*args):
    """return IsADirectoryError"""

    return OSError(errno.EISDIR, strerror(errno.EISDIR), *args)


class StatResult(NamedTuple):
    st_mode: int
    st_ino: int
    st_dev: int = 0
    st_nlink: Optional[int] = None
    st_uid: Optional[int] = None
    st_gid: Optional[int] = None
    st_size: Optional[int] = None
    st_atime: Optional[int] = None
    st_mtime: Optional[int] = None
    st_ctime: Optional[int] = None
    st_atime_ns: Optional[int] = None
    st_mtime_ns: Optional[int] = None
    st_ctime_ns: Optional[int] = None


class MemoryPurePath:
    def joinpath(self, *pathsegments: str) -> Self:
        raise NotImplementedError


class MemoryPath(MemoryPurePath):
    __slots__ = ("_data", "_children")

    _data: Optional[bytes]
    children: "Optional[List[MemoryPath]]"

    def __init__(self, data: Optional[bytes] = None, children: "Optional[List[MemoryPath]]" = None) -> None:
        if data is not None and children is not None:
            raise ValueError("A path cannot have data and children")

        self._data = data
        self._children = children

    # Parsing and generating URIs

    @classmethod
    def from_uri(cls, uri: str) -> Self:
        raise NotImplementedError

    def as_uri(self) -> str:
        raise NotImplementedError

    # Expanding and resolving paths

    @classmethod
    def home(cls) -> Self:
        raise NotImplementedError

    def expanduser(self) -> Self:
        raise NotImplementedError

    @classmethod
    def cwd(cls) -> Self:
        raise NotImplementedError

    def absolute(self) -> Self:
        raise NotImplementedError

    def resolve(self, strict: bool = False) -> Self:
        raise NotImplementedError

    def readlink(self) -> Self:
        raise NotImplementedError

    # Querying file type and status

    def stat(self, *, follow_symlinks: bool = True) -> StatResult:
        if self._data is None and self._children is None:
            raise file_not_found_error(str(self))

        if self._children is not None:
            st_mode: int = stat.S_IFDIR
            st_ino = id(self)
            st_size = None
        else:
            assert self._data is not None  # for mypy
            st_mode = stat.S_IFREG
            st_ino = id(self)
            st_size = len(self._data)

        return StatResult(st_mode=st_mode, st_ino=st_ino, st_size=st_size)

    def lstat(self) -> os.stat_result:
        raise NotImplementedError

    def exists(self, *, follow_symlinks: bool = True) -> bool:
        return self._data is not None or self._children is not None

    def is_file(self, *, follow_symlinks: bool = True) -> bool:
        return self._data is not None

    def is_dir(self, *, follow_symlinks: bool = True) -> bool:
        return self._children is not None

    def is_symlink(self) -> bool:
        return False

    def is_junction(self) -> bool:
        return False

    def is_mount(self) -> bool:
        return False

    def is_socket(self) -> bool:
        return False

    def is_fifo(self) -> bool:
        return False

    def is_block_device(self) -> bool:
        return False

    def is_char_device(self) -> bool:
        return False

    def samefile(self, other: "MemoryPath") -> bool:
        return self is other

    # Reading and writing files

    @contextmanager
    def open(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
    ) -> Iterator[IO]:
        if "w" not in mode and self._data is None:
            raise file_not_found_error(str(self))

        if "b" in mode:
            if encoding is not None or errors is not None or newline is not None:
                raise ValueError("encoding, errors and newline cannot by used for bytes")

            if self._data is None:
                bdata = b""
            else:
                bdata = self._data

            with BytesIO(bdata) as fp:
                yield fp
                self._data = fp.getvalue()
        else:
            if encoding is None:
                encoding = DEFAULT_ENCODING

            if errors is None:
                errors = "strict"

            if self._data is None:
                sdata = ""
            else:
                sdata = self._data.decode(encoding, errors)

            with StringIO(sdata, newline) as fp:
                yield fp
                self._data = fp.getvalue().encode(encoding, errors)

    def read_text(
        self, encoding: Optional[str] = None, errors: Optional[str] = None, newline: Optional[str] = None
    ) -> str:
        # new line param is ignored

        if self._children is not None:
            raise is_a_directory_error(str(self))

        if self._data is None:
            raise file_not_found_error(str(self))

        if encoding is None:
            encoding = DEFAULT_ENCODING

        if errors is None:
            errors = "strict"

        return self._data.decode(encoding, errors)

    def read_bytes(self) -> bytes:
        if self._children is not None:
            raise is_a_directory_error(str(self))

        if self._data is None:
            raise file_not_found_error(str(self))

        return self._data

    def write_text(self, data: str, encoding=None, errors=None, newline=None) -> None:
        # new line param is ignored

        if self._children is not None:
            raise is_a_directory_error(str(self))

        if encoding is None:
            encoding = DEFAULT_ENCODING

        if errors is None:
            errors = "strict"

        self._data = data.encode(encoding, errors)

    def write_bytes(self, data: bytes) -> None:
        if self._children is not None:
            raise is_a_directory_error(str(self))

        self._data = data

    # Reading directories

    def iterdir(self) -> "Iterator[MemoryPath]":
        if self._data is not None:
            raise not_a_directory_error(str(self))

        if self._children is None:
            raise file_not_found_error(str(self))

        yield from self._children

    def glob(
        self, pattern, *, case_sensitive: Optional[bool] = None, recurse_symlinks: bool = False
    ) -> "Iterator[MemoryPath]":
        raise NotImplementedError

    def rglob(
        self, pattern, *, case_sensitive: Optional[bool] = None, recurse_symlinks: bool = False
    ) -> "Iterator[MemoryPath]":
        raise NotImplementedError

    def walk(
        self, top_down=True, on_error=None, follow_symlinks: bool = False
    ) -> "Iterator[Tuple[MemoryPath, List[str], List[str]]]":
        raise NotImplementedError

    # Creating files and directories

    def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
        if self._children is not None:
            raise file_exists_error(str(self))

        if self._data is None:
            self._data = b""
        elif not exist_ok:
            raise file_exists_error(str(self))

    def mkdir(self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False) -> None:
        if self._data is not None:
            raise file_exists_error(str(self))

        if self._children is None:
            self._children = []
        elif not exist_ok:
            raise file_exists_error(str(self))

    def symlink_to(self, target: str, target_is_directory: bool = False) -> None:
        raise NotImplementedError

    def hardlink_to(self, target: str) -> None:
        raise NotImplementedError



import numpy as np
import polars as pl


def pl_index(df: pl.DataFrame, indices: np.ndarray, index_col: str = "index") -> pl.DataFrame:
    assert index_col not in df.columns

    return (
        df.with_row_index(index_col)
        .join(
            pl.DataFrame({index_col: indices}),
            on=index_col,
            how="inner",
            validate="1:1",
            maintain_order="left",
        )
        .drop(index_col)
    )



from collections import defaultdict
from contextlib import nullcontext
from difflib import unified_diff
from functools import wraps
from itertools import chain, product, zip_longest
from typing import IO, Any, Callable, Collection, DefaultDict, Hashable, Iterable, Mapping, Optional, Tuple, TypeVar
from unittest import TestCase
from unittest.result import TestResult
from unittest.runner import TextTestRunner

from typing_extensions import ParamSpec

from .file import equal_files
from .iter import progress
from .typing import ExceptionsType

T = TypeVar("T")
U = TypeVar("U")
P = ParamSpec("P")

anullcontext = nullcontext()


def print_file_diff(fromfile: str, tofile: str, encoding: str) -> None:
    with open(fromfile, encoding=encoding) as fr:
        fromlines = fr.readlines()
    with open(tofile, encoding=encoding) as fr:
        tolines = fr.readlines()

    for line in unified_diff(fromlines, tolines, fromfile, tofile):
        print(line, end="")


class MyTestResult(TestResult):
    """Extension of TestResult to support numbering test cases"""

    def __init__(self, stream: IO[str], descriptions, verbosity):
        """Initializes the test number generator, then calls super impl"""
        self.stream = stream
        self.current_test = 0
        self.current_subtest = 0
        TestResult.__init__(self, stream, descriptions, verbosity)

    def progress(self) -> None:
        self.stream.write(f"Running test #{self.current_test}, subtest #{self.current_subtest}\r")

    def startTest(self, test) -> None:
        self.current_test += 1
        TestResult.startTest(self, test)
        self.progress()

    def stopTest(self, test) -> None:
        TestResult.stopTest(self, test)

    def stopTestRun(self) -> None:
        self.stream.write("\n")
        TestResult.stopTestRun(self)

    def addSubTest(self, test, subtest, outcome) -> None:
        """Called when a subtest finishes."""

        TestResult.addSubTest(self, test, subtest, outcome)
        self.testsRun += 1
        self.current_subtest += 1
        self.progress()


class MyTestRunner(TextTestRunner):
    """Extension of TestRunner to support numbering test cases"""

    resultclass = MyTestResult

    def run(self, test):
        """Stores the total count of test cases, then calls super impl"""

        self.total_test_cases = test.countTestCases()
        return TextTestRunner.run(self, test)

    def _makeResult(self):
        return TextTestRunner._makeResult(self)


class NoRaise:
    def __init__(self, testcase: TestCase, message: Optional[str] = None) -> None:
        self.testcase = testcase
        self.message = message

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            if self.message:
                self.testcase.fail(self.message)
            else:
                self.testcase.fail(exc_value)


def make_comparable(d: Any) -> Any:
    if isinstance(d, (str, int, type(None))):
        return d
    elif isinstance(d, list):
        return list(make_comparable(i) for i in d)
    elif isinstance(d, tuple):
        return tuple(make_comparable(i) for i in d)
    elif isinstance(d, dict):
        return make_comparable(list(d.items()))
    else:
        raise TypeError(f"must be list, tuple or dict, not {type(d)}")


def is_equal_unordered(seq_a: Collection, seq_b: Collection) -> bool:
    if isinstance(seq_a, set) and isinstance(seq_b, set):
        return seq_a == seq_b

    seq_a = sorted(make_comparable(seq_a))
    seq_b = sorted(make_comparable(seq_b))
    return seq_a == seq_b


class MyTestCase(TestCase):
    def assertAnd(self, first: Any, second: Any, msg: Optional[str] = None) -> None:
        self.assertTrue(first and second, msg)

    def assertTypeEqual(self, first: Any, second: Any, msg: Optional[str] = None) -> None:
        self.assertEqual(type(first), type(second), msg)

    def assertFilesEqual(self, first_path: Any, second_path: Any, msg: Optional[str] = None) -> None:
        self.assertTrue(equal_files(first_path, second_path), msg)

    def assertIterEqual(self, first: Iterable, second: Iterable, msg: Optional[str] = None) -> None:
        for i, (a, b) in enumerate(zip_longest(first, second)):
            if msg:
                msg = " : " + str(msg)
            self.assertEqual(a, b, msg=f"in iteration index {i}: {msg}")

    def assertIterAlmostEqual(self, first: Iterable, second: Iterable, msg: Optional[str] = None) -> None:
        for i, (a, b) in enumerate(zip_longest(first, second)):
            if msg:
                msg = " : " + str(msg)
            try:
                self.assertAlmostEqual(a, b, msg=f"in iteration index {i}: {msg}")
            except TypeError:
                raise AssertionError("Invalid types (probably different length iters)")  # from None

    def assertAllEqual(self, args: Iterable, msg: Optional[str] = None) -> None:  # *args doesn't work in python2!?
        it = iter(args)
        first = next(it)
        for second in it:
            self.assertEqual(first, second, msg)

    def assertIterIterEqual(
        self, first: Iterable[Iterable], second: Iterable[Iterable], msg: Optional[str] = None
    ) -> None:  # UNTESTED
        for i, (a, b) in enumerate(zip_longest(first, second)):
            if msg:
                msg = " : " + str(msg)
            self.assertIterEqual(a, b, msg=f"in iteration index {i}: {msg}")

    def assertPriorityEqual(
        self,
        first: Iterable[Optional[Tuple[Hashable, Hashable]]],
        second: Iterable[Optional[Tuple[Hashable, Hashable]]],
        msg: Optional[str] = None,
    ) -> None:
        d1: DefaultDict[Hashable, set] = defaultdict(set)
        d2: DefaultDict[Hashable, set] = defaultdict(set)

        for a, b in zip_longest(first, second):
            if a is not None:
                val_1, p_1 = a
                if b is None and p_1 == 0:
                    continue

            if b is not None:
                val_2, p_2 = b
                if a is None and p_2 == 0:
                    continue

            self.assertAnd(a, b)

            if val_1 == 0 and val_2 == 0:
                continue

            d1[p_1].add(val_1)
            d2[p_2].add(val_2)
        self.assertEqual(d1, d2, msg)

    def assertUnorderedSeqEqual(self, first: Iterable, second: Iterable, msg: Optional[str] = None) -> None:
        # fixme: use is_equal_unordered

        first = sorted(first)
        second = sorted(second)
        self.assertEqual(first, second, msg)

    def assertNoRaise(self, msg: Optional[str] = None) -> NoRaise:
        return NoRaise(self, msg)

    def assertEqualsOneOf(self, result: T, truths: Iterable[T]) -> None:
        for truth in truths:
            if result == truth:
                return

        raise AssertionError("Result does not match one of the provided truths")  # from None

    def assertUnorderedMappingEqual(
        self, first: Mapping[T, U], second: Mapping[T, U], msg: Optional[str] = None
    ) -> None:
        self.assertEqual(len(first), len(second), msg)

        for a, b in zip(sorted(first.keys()), sorted(second.keys())):
            self.assertEqual(a, b)
            self.assertEqual(first[a], second[b])


def random_arguments(n: int, *funcs: Callable[[], Any]) -> Callable[[Callable], Callable]:
    def decorator(func):
        @wraps(func)
        def inner(self):
            for i in range(n):
                with self.subTest(str(i)):
                    if func(self, *(f() for f in funcs)) is not None:
                        raise AssertionError

        return inner

    return decorator


# also called: parameterize
def parametrize(*args_list: tuple) -> Callable[[Callable], Callable]:
    def decorator(func):
        @wraps(func)
        def inner(self):
            for args in args_list:
                with self.subTest(str(args)[:1000]):
                    if func(self, *args) is not None:
                        raise AssertionError

        return inner

    return decorator


def parametrize_product(*args_list: Iterable) -> Callable[[Callable], Callable]:
    def decorator(func):
        @wraps(func)
        def inner(self):
            for args in product(*args_list):
                with self.subTest(str(args)):
                    if func(self, *args) is not None:
                        raise AssertionError

        return inner

    return decorator


def parametrize_starproduct(*args_list: Iterable[Iterable]) -> Callable[[Callable], Callable]:
    def decorator(func):
        @wraps(func)
        def inner(self):
            for args in product(*args_list):
                combined_args = tuple(chain.from_iterable(args))
                with self.subTest(str(combined_args)):
                    if func(self, *combined_args) is not None:
                        raise AssertionError

        return inner

    return decorator


def repeat(number: int, verbose: bool = False) -> Callable[[Callable[P, Any]], Callable[P, None]]:
    def decorator(func: Callable[P, Any]) -> Callable[P, None]:
        @wraps(func)
        def inner(self, *args: P.args, **kwargs: P.kwargs) -> None:
            for _i in progress(range(number), disable=not verbose):
                if func(self, *args, **kwargs) is not None:  # no self.subTest(str(i))
                    raise AssertionError

        return inner

    return decorator


def skip_on_exception(exceptions: ExceptionsType) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def inner(self, *args: P.args, **kwargs: P.kwargs) -> T:
            try:
                return func(self, *args, **kwargs)
            except exceptions as e:
                self.skipTest(f"Skipping: {type(e).__name__}")

        return inner

    return decorator



from functools import wraps
from typing import Callable, TypeVar
from warnings import warn

from typing_extensions import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")


def deprecated(msg: str, stacklevel: int = 2) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def inner(*args: P.args, **kwargs: P.kwargs) -> T:
            warn(msg, DeprecationWarning, stacklevel)  # noqa: B028
            print("DeprecationWarning:", msg)
            return func(*args, **kwargs)

        return inner

    return decorator



import logging
import os
import os.path
from datetime import timedelta
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, Optional, TypeVar

from typing_extensions import ParamSpec

from .datetime import now
from .filesystem import mdatetime
from .object import args_to_key
from .time import MeasureTime

T = TypeVar("T")
P = ParamSpec("P")

logger = logging.getLogger(__name__)


def cache(
    path: Path,
    duration: Optional[timedelta] = None,
    generator: bool = False,
    serializer: str = "pickle",
    serializer_kwargs: Optional[Dict[str, Any]] = None,
    deserializer_kwargs: Optional[Dict[str, Any]] = None,
    ignoreargs: bool = False,
    ignore_first_arg: bool = False,
    keyfunc: Optional[Callable[[tuple, dict], str]] = None,
    consume: bool = False,
    return_cached: bool = False,
    file_ext: Optional[str] = None,
    cached_only: bool = False,
) -> Callable[[Callable[P, T]], Callable[P, Any]]:
    """Decorator to cache function calls. Doesn't take function arguments into regard.
    It's using `pickle` to deserialize the data. So don't use it with untrusted inputs.

    `path`: path to cache file or directory.
            This can be a template string where `ppv` is assigned the pickle protocol version.
    `duration`: maximum age of cache
    `generator`: set to True to store the results of generator objects
    `protocol`: pickle protocol version
    `file_ext`: File extension to use for cache files. Should start with `.`.

    If `ignoreargs` is True, the cache won't take function arguments into regard.
            The path will be interpreted as template string to a file instead of a directory.
    """

    from . import pickle  # nosec

    serializer_kwargs = serializer_kwargs or {}
    deserializer_kwargs = deserializer_kwargs or {}

    if serializer == "pickle":
        from pickle import HIGHEST_PROTOCOL  # nosec

        _serializer_kwargs: Dict[str, Any] = {"protocol": HIGHEST_PROTOCOL, "safe": True}
        _deserializer_kwargs: Dict[str, Any] = {}
        write_iter = pickle.write_iter
        write_file = pickle.write_pickle
        read_iter = pickle.read_iter
        read_file = pickle.read_pickle
        key_to_hash = pickle.key_to_hash
        _hash_sep: Any = None
        _file_ext = file_ext or ".p"

    elif serializer == "msgpack":
        from . import msgpack

        _serializer_kwargs = {"safe": True}
        _deserializer_kwargs = {}
        write_iter = msgpack.write_iter
        write_file = msgpack.write_msgpack
        read_iter = msgpack.read_iter
        read_file = msgpack.read_msgpack
        key_to_hash = msgpack.key_to_hash
        _hash_sep = None
        _file_ext = file_ext or ".msgpack"

    elif serializer == "json":
        from . import json

        _serializer_kwargs = {"ensure_ascii": False, "sort_keys": False, "default": None, "safe": True}
        _deserializer_kwargs = {"cls": None, "object_hook": None}
        write_iter = json.write_json_lines
        write_file = json.write_json
        read_iter = json.read_json_lines
        read_file = json.read_json
        key_to_hash = json.key_to_hash
        _hash_sep = {}
        _file_ext = file_ext or ".json"

    else:
        raise ValueError(f"Invalid serializer: {serializer}")

    _serializer_kwargs.update(serializer_kwargs)
    _deserializer_kwargs.update(deserializer_kwargs)

    _pure_serializer_kwargs = _serializer_kwargs.copy()
    _pure_serializer_kwargs.pop("safe")

    if not generator and consume:
        raise ValueError("consume can only be used for generator")

    if duration is None:
        _duration = timedelta.max
    else:
        _duration = duration

    def decorator(func: Callable[P, T]) -> Callable[P, Any]:
        @wraps(func)
        def inner(*args: P.args, **kwargs: P.kwargs) -> Any:
            strpath = os.fspath(path).format_map(_serializer_kwargs)

            if not ignoreargs:
                if ignore_first_arg:
                    args = args[1:]

                if keyfunc is None:
                    hashstr = key_to_hash(args_to_key(args, kwargs, _hash_sep), **_pure_serializer_kwargs)
                else:
                    hashstr = keyfunc(args, kwargs)
                fullpath = os.path.join(strpath, hashstr + _file_ext)
            else:
                if keyfunc or file_ext:
                    raise ValueError("`keyfunc` or `file_ext` can only be specified if ignoreargs is False")
                if args or kwargs:
                    logger.warning("cache file decorator for %s called with arguments", func.__name__)
                fullpath = strpath

            try:
                invalid = now() - mdatetime(fullpath) > _duration
            except FileNotFoundError:
                invalid = True

            if invalid:
                if cached_only:
                    raise LookupError("Not in cache")

                cached = False
                if not ignoreargs:
                    path.mkdir(parents=True, exist_ok=True)

                if generator and not consume:
                    it = func(*args, **kwargs)
                    logger.debug("Writing iterable to <%s>", fullpath)
                    result: Any = write_iter(it, fullpath, **_serializer_kwargs)
                else:
                    with MeasureTime() as m:
                        if generator and consume:
                            result = list(func(*args, **kwargs))
                        else:
                            result = func(*args, **kwargs)
                    logger.debug("Result calculated in %s seconds and written to <%s>", m.get(), fullpath)
                    write_file(result, fullpath, **_serializer_kwargs)
            else:
                cached = True
                if generator and not consume:
                    logger.debug("Loading iterable from <%s>", fullpath)
                    result = read_iter(fullpath, **_deserializer_kwargs)
                else:
                    with MeasureTime() as m:
                        result = read_file(fullpath, **_deserializer_kwargs)
                    logger.debug("Result loaded from <%s> in %s seconds", fullpath, m.get())

            if return_cached:
                return cached, result
            else:
                return result

        return inner

    return decorator



from typing import Any

import datasets
import evaluate
from sklearn.metrics import accuracy_score


class Accuracy(evaluate.Metric):
    def _info(self):
        return evaluate.MetricInfo(
            description="Binary, multi-class or Multi-label accuracy",
            citation="",
            inputs_description="A dataset with `predictions` and `references` columns",
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("int32")),
                    "references": datasets.Sequence(datasets.Value("int32")),
                }
                if self.config_name == "multilabel"
                else {
                    "predictions": datasets.Value("int32"),
                    "references": datasets.Value("int32"),
                }
            ),
            reference_urls=["https://scikit-learn.org/stable/modules/generated/sklearn.metrics.accuracy_score.html"],
        )

    def _compute(
        self,
        predictions,
        references,
        normalize=True,
        sample_weight=None,
        pos_label: Any = None,
        average: Any = None,
        zero_division: Any = None,
    ):
        """pos_label, average and zero_division are ignored"""

        return {
            "accuracy": float(accuracy_score(references, predictions, normalize=normalize, sample_weight=sample_weight))
        }



import json
from itertools import chain

from setuptools import setup

# nltk.download("punkt")

with open("install_requires.json", encoding="utf-8") as fr:
    install_requires = json.load(fr)

with open("extras_require.json", encoding="utf-8") as fr:
    extras_require = json.load(fr)

extras_require["all"] = sorted(set(chain.from_iterable(extras_require.values())))

setup(
    packages=["genutility", "genutility.compat", "genutility.fileformats", "genutility.win"],
    package_data={"genutility": ["py.typed", "fileformats/data/*.tsv"]},
    install_requires=install_requires,
    extras_require=extras_require,
)



from math import exp, sqrt
from typing import Any, Callable, Dict, Generic, Iterator, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np

# from .numba import opjit

T = TypeVar("T")

RGB_YELLOW = (255, 255, 0)
RGB_WHITE = (255, 255, 255)


def assert_square(name: str, value: np.ndarray) -> None:
    """Raises a ValueError if matrix `value` is not square.
    value: [A, A]
    """

    if not len(value.shape) == 2 or value.shape[0] != value.shape[1]:
        raise ValueError(f"{name} must be a square")


def normalized_choice(
    p_ind, p_prob
):  # is this neccessary? I think `np.random.choice` can handle unnormalized probabilities
    return np.random.choice(p_ind, p=p_prob / np.sum(p_prob))


def shannon_entropy(ps: np.ndarray, base: int = 2) -> float:
    """Calculates the Shannon entropy for probabilities `ps` with `base`."""

    return -np.sum(ps * np.log(ps) / np.log(base))


def is_rgb(img: np.ndarray) -> bool:
    """Simply tests if `img` has 3 channels."""

    return len(img.shape) >= 1 and img.shape[-1] == 3


# @opjit() # np.errstate doesn't work in numba...
def rgb_to_hsi(image: np.ndarray) -> np.ndarray:
    """Converts an array [..., channels] of RGB values to HSI color values (H in rad).
    RGB values are assumed to be normalized to (0, 1).
    """

    if not is_rgb(image):
        raise ValueError("Input needs to be an array of RGB values")

    r = image[..., 0]
    g = image[..., 1]
    b = image[..., 2]

    out = np.zeros_like(image)

    # allequal = (img == img[:, :, 0, np.newaxis]).all(axis=-1)

    with np.errstate(invalid="ignore"):
        tmp = (2.0 * r - g - b) / 2.0 / np.sqrt((r - g) ** 2 + (r - b) * (g - b))  # if r==g==b then 0/0

        theta = np.arccos(np.clip(tmp, -1.0, +1.0))

        out[..., 0] = np.where(b <= g, theta, 2 * np.pi - theta)  # H
        out[..., 2] = np.sum(image, axis=-1) / 3.0  # I
        out[..., 1] = 1 - np.amin(image, axis=-1) / out[..., 2]  # S if r==g==b==0 then 0/0

    np.nan_to_num(out[..., 0:2], copy=False)

    return out


# @opjit() # np.dot with more than 2 dimensions is not supported...
def rgb_to_ycbcr(image: np.ndarray) -> np.ndarray:
    """Converts an array [..., channels] of RGB values to Digital Y'CbCr (0-255).
    RGB values are assumed to be normalized to (0, 1).
    Don't forget to cast to uint8 for pillow.
    """

    """  from RGB (0-1).
    """

    if not is_rgb(image):
        raise ValueError("Input needs to be an array of RGB values")

    m = np.array(
        [
            [+065.481, +128.553, +024.966],
            [-037.797, -074.203, +112.000],
            [+112.000, -093.786, -018.214],
        ]
    )
    a = np.array([16, 128, 128])

    return np.dot(image, m.T) + a


def random_triangular_matrix(size: int, lower: bool = True) -> np.ndarray:
    """Returns a triangular matrix with random value between 0 and 1 uniformly."""

    a = np.random.uniform(0, 1, (size, size))
    if lower:
        ind = np.triu_indices(5, 1)
    else:
        ind = np.tril_indices(5, 1)
    a[ind] = 0

    return a


def is_square(A: np.ndarray) -> bool:
    if len(A.shape) != 2:
        return False

    if A.shape[0] != A.shape[1]:
        return False

    return True


def batch_vTAv(A: np.ndarray, v: np.ndarray) -> np.ndarray:
    """Performs batched calculation of `v^T A v` transform.
    Special case of bilinear form `x^T A y`

    A: float[B, X, X]
    v: float[B, X]

    returns: float[B]
    """

    """ Faster than
        Av = np.matmul(A, v[...,:,None]) # [B, X, 1]
        return np.matmul(v[...,None,:], Av).squeeze((-2, -1)) # [B]
    """

    return np.einsum("...k,...kl,...l->...", v, A, v)


def batch_inner(a: np.ndarray, b: np.ndarray, verify: bool = True) -> np.ndarray:
    """Performs a batched inner product over the last dimension.
    Replacement for deprecated `from numpy.core.umath_tests import inner1d`.
    Shapes: (B, X), (B, X) -> (B, )
    """

    if verify and a.shape != b.shape:
        raise ValueError("All dimensions have to be equal")

    if a.shape[-1] == 0:
        return np.empty_like(a)

    return np.einsum("...i,...i->...", a, b)  # faster than np.sum(a * b, axis=-1)


def batch_outer(a: np.ndarray, b: np.ndarray, verify: bool = True) -> np.ndarray:
    """Performs a batched outer product over the last dimension.
    Shapes: (B, X), (B, Y) -> (B, X, Y)
    """

    if verify and a.shape[:-1] != b.shape[:-1]:
        raise ValueError("All except the last dimension have to be equal")

    return np.einsum("...i,...j->...ij", a, b)  # slightly faster than np.multiply(a[...,:,None], b[...,None,:])


def batchtopk(
    probs: np.ndarray, k: Optional[int] = None, axis: int = -1, reverse: bool = False
) -> Tuple[np.ndarray, np.ndarray]:
    """`probs` values ndarray
    `k` take the smallest `k` elements, if `reverse` is False
            and the largest `k` if `reverse` is True
    `axis` sorting and selection axis.
    """

    if k is not None and k <= 0:
        raise ValueError("k must be larger than 0. Use None to chose all elements.")

    if axis != -1:
        raise ValueError("Only last axis supported atm")

    if len(probs.shape) <= 1:
        raise ValueError("probs must be at least 2-dimensional")

    if reverse:
        sign = -1
    else:
        sign = 1

    indices = np.argsort(sign * probs, axis=-1)  # use argpartition?
    probs = np.take_along_axis(probs, indices[..., :k], axis=-1)

    return indices, probs


# @opjit() # np.diagonal not supported by numba as of version 0.52.0
def logtrace(m: np.ndarray) -> np.ndarray:
    """Calcuates the sum of the logs of the diagonal elements (batchwise if necessary)
    m: [..., x, x]
    """

    """ note: performance cannot easily be improve by numba.
        `np.diagonal` not supported by numba 0.52.0
    """

    return np.sum(np.log(np.diagonal(m, axis1=-2, axis2=-1)), axis=-1)


def shiftedexp(pvals: np.ndarray) -> np.ndarray:
    """Shifts `pvals` by the largest value in the last dimension
    before the exp is calculated to prevent overflow (batchwise if necessary).
    Can be used if probabilities are normalized again later.
    """

    if pvals.shape[-1] == 0:
        return np.empty_like(pvals)

    return np.exp(pvals - np.amax(pvals, axis=-1)[..., None])


class Sampler:
    """Sample from discrete CDF."""

    def __init__(self, cdf: np.ndarray) -> None:
        self.cdf = cdf
        self.psum = cdf[-1]

    def __call__(self) -> int:
        """Sample one."""

        rand = np.random.uniform(0, self.psum)
        return np.searchsorted(self.cdf, rand, side="right")

    def sample(self, n: int) -> np.ndarray:
        """Sample `n`."""

        rands = np.random.uniform(0, self.psum, n)
        return np.searchsorted(self.cdf, rands, side="right")

    def pdf(self, n: int, minlength: Optional[int] = None) -> np.ndarray:
        out = np.bincount(self.sample(n), minlength=minlength)
        return out / n


def sample_probabilities(pvals: np.ndarray) -> Callable[[], int]:
    """Sample from list of probabilities `pvals` with replacement.
    The probabilities don't need to be normalized.
    """

    return Sampler(np.cumsum(pvals))


class UnboundedSparseMatrix(Generic[T]):
    def __init__(self, dtype: type = float) -> None:
        self.dtype = dtype
        self.zero = self.dtype(0)
        self.m: Dict[Tuple[int, int], T] = {}
        self.cols = 0
        self.rows = 0

    def __getitem__(self, slice: Tuple[int, int]) -> T:
        return self.m.get(slice, self.zero)

    def __setitem__(self, slice: Tuple[int, int], value: T) -> None:
        c, r = slice
        self.cols = max(self.cols, c + 1)
        self.rows = max(self.rows, r + 1)
        self.m[slice] = value

    def todense(self) -> np.ndarray:
        ret = np.zeros((self.cols, self.rows), dtype=self.dtype)

        for slice, value in self.m.items():
            ret[slice] = value

        return ret


def normalize(pvals: np.ndarray) -> np.ndarray:
    return pvals / np.sum(pvals)


def categorical(pvals: np.ndarray) -> int:
    """Sample from the categorical distribution using `pvals`.
    See: https://en.wikipedia.org/wiki/Categorical_distribution
    """

    return sample_probabilities(pvals)()  # faster than: np.argmax(np.random.multinomial(1, normalize(pvals)))


def population2cdf(population: np.ndarray) -> np.ndarray:
    """Convert a population (list of observations) to a CDF."""

    population = np.sort(population)
    return np.searchsorted(population, population, side="right") / len(population)


def pmf2cdf(pdf: np.ndarray) -> np.ndarray:
    """Convert a discrete PDF into a discrete CDF."""

    cdf = np.cumsum(pdf)
    return cdf / cdf[-1]


def _two_sample_kolmogorov_smirnov_same_length(
    cdf1: np.ndarray, cdf2: np.ndarray, n1: int, n2: int
) -> Tuple[float, float]:
    # note: yields different results as `scipy.stats.ks_2samp`

    if len(cdf1) != len(cdf2):
        raise ValueError("Both CDFs must have the same length")

    D = np.amax(np.abs(cdf1 - cdf2))  # K-S statistic
    level = exp(-2 * (D / sqrt((n1 + n2) / (n1 * n2))) ** 2)
    return D, level


def _two_sample_kolmogorov_smirnov_population(
    p1: np.ndarray, p2: np.ndarray, alpha: float = 0.05
) -> Tuple[float, float, bool]:
    # note: yields different results as `scipy.stats.ks_2samp`

    cdf1 = population2cdf(p1)
    cdf2 = population2cdf(p2)

    statistic, pvalue = _two_sample_kolmogorov_smirnov_same_length(cdf1, cdf2, len(cdf1), len(cdf2))
    reject = pvalue < alpha
    return statistic, pvalue, reject


def _two_sample_kolmogorov_smirnov_pmf(
    pmf1: np.ndarray, pmf2: np.ndarray, alpha: float = 0.05
) -> Tuple[float, float, bool]:
    # note: yields different results as `scipy.stats.ks_2samp`
    """Tests the null hypothesis that both samples belong to the same distribution.
    See: https://en.wikipedia.org/wiki/Kolmogorov%E2%80%93Smirnov_test#Two-sample_Kolmogorov%E2%80%93Smirnov_test
    """

    cdf1 = np.cumsum(pmf1)
    cdf2 = np.cumsum(pmf2)

    n1 = cdf1[-1]
    n2 = cdf2[-1]

    # cannot be inplace because of type conversion
    cdf1 = cdf1 / n1
    cdf2 = cdf2 / n2

    statistic, pvalue = _two_sample_kolmogorov_smirnov_same_length(cdf1, cdf2, n1, n2)
    reject = pvalue < alpha
    return statistic, pvalue, reject


def inf_matrix_power(pm: np.ndarray, dtype=np.float64) -> np.ndarray:
    """Calculate stochastic matrix `pm` to the power of infinity,
    by finding the eigenvector which corresponds to the eigenvalue 1.
    """

    w, v = np.linalg.eig(
        pm
    )  # scipy.linalg.eig would probably by faster as it can return the left and right eigen vectors

    if not np.isclose(w[0], 1.0):
        raise ValueError("The first eigenvalue is not none. Is this a right stochastic matrix?")

    vi = np.linalg.inv(v)
    d = np.zeros(pm.shape[0], dtype=dtype)
    d[0] = 1.0

    return np.matmul(v, np.matmul(np.diag(d), vi))


def decompress(selectors: np.ndarray, data: np.ndarray, default) -> np.ndarray:
    """Same result as:
    `np.array(list(genutility.iter.decompress(selectors, iter(data), default)))`
    but faster.
    `selectors` should be a bool array.
    """

    out = np.full(len(selectors), default)
    out[selectors] = data
    return out


def unblock(arr: np.ndarray, n1: int, n2: int, axis1: int = -1, axis2: int = -2, blocksize: bool = False) -> np.ndarray:
    """Inverse of np.block.
    Set axis to (-2, -1) to modify the order of the result.
    """

    """ test (stackoverflow): Ok, so considering I have N block matrices with bm x bn dimension and want to stack them in a m x n matrix, provided N = m x n, I would then have x.reshape(m,n,bm,bn).swapaxes(1,2).reshape(bm*m,-1)
    """

    s = np.array(arr.shape)
    if s[axis1] % n1 != 0 or s[axis2] % n2 != 0:
        raise ValueError(f"{s[axis1]}x{s[axis2]} does not divide by {n1}x{n2}")

    if blocksize:
        n1 = s[axis1] // n1
        n2 = s[axis2] // n2

    # this first .split adds a new dimensions on the outside, so if a absolute index
    # is given for the second axis it must be moved one to the right
    if axis2 >= 0:
        _axis2 = axis2 + 1
    else:
        _axis2 = axis2

    arr = np.array(np.split(arr, n1, axis1))
    arr = np.array(np.split(arr, n2, _axis2))

    inv_blocksize = n1 * n2
    total = s[axis1] * s[axis2]
    s[axis2] = inv_blocksize
    s[axis1] = total // inv_blocksize

    return np.reshape(arr, s)


def unblock2d(arr: np.ndarray, n1: int, n2: int, blocksize: bool = False) -> np.ndarray:
    """
    arr: [width, height, ...]

    returns: (block_num, block_size)
    """

    s = np.array(arr.shape)
    if blocksize:
        bs1 = n1
        bs2 = n2
        bn1 = s[0] // n1
        bn2 = s[1] // n2
    else:
        bn1 = n1
        bn2 = n2
        bs1 = s[0] // n1
        bs2 = s[1] // n2

    if len(arr.shape) > 2:
        return arr.reshape(bn1, bs1, bn2, bs2, -1).swapaxes(1, 2).reshape(bn1 * bn2, bs1 * bs2, -1)
    else:
        return arr.reshape(bn1, bs1, bn2, bs2).swapaxes(1, 2).reshape(bn1 * bn2, bs1 * bs2)


def block2d(arr: np.ndarray, bn1: int, bn2: int, bs1: int, bs2: int) -> np.ndarray:
    """
    arr: (block_num, block_size)

    returns: (width, height)
    """

    s = np.array(arr.shape)
    assert s[0] == bn1 * bn2
    assert s[1] == bs1 * bs2

    return arr.reshape(bn1, bn2, bs1, bs2).swapaxes(1, 2).reshape(bn1 * bs1, bn2 * bs2)


def remove_color(img: np.ndarray, ratio: float, neutral_color: Tuple[int, int, int] = RGB_WHITE) -> None:
    """Replace colored pixels with a `neutral_color`. The `ratio` defines the 'colorfulness' above
    which level the pixel should be replace.
    I.e. if the `ratio` is 1 nothing will be replaced,
    if `ratio` is 0 only strict greys are kept unmodified.
    """

    channels = img.shape[-1]
    assert channels == 3, "Not a 3 channel color image"

    norm = np.std(np.array(RGB_YELLOW))  # this is the same for all pure colors

    sd = np.std(img, axis=-1)
    img[sd > ratio * norm] = neutral_color


# @opjit(cache=False) doesn't help a lot, because it's a generator
def sliding_window_2d(
    image: np.ndarray, window_size: Tuple[int, int], step_size: Tuple[int, int] = (1, 1)
) -> Iterator[np.ndarray]:
    win_x, win_y = window_size
    step_x, step_y = step_size
    height, width = image.shape[0:2]

    for y in range(0, height - win_y + 1, step_y):
        for x in range(0, width - win_x + 1, step_x):
            yield image[y : y + win_y, x : x + win_x, ...]


def histogram_correlation(hist1: np.ndarray, hist2: np.ndarray) -> np.ndarray:
    """Input shape of `hist1` and `hist2`: [batch..., levels]. The correlation is calculated over `levels`
    and then batched over the remaining dimensions.
    Similar to `cv2.compareHist(hist1, hist2, cv2.HISTCMP_CORREL)`.
    """

    if hist1.ndim < 1 or hist1.shape[-1] < 1 or hist1.shape != hist2.shape:
        raise ValueError(f"Histogram shapes unsupported: {hist1.shape}, {hist2.shape}")

    h1norm = (hist1.T - np.mean(hist1, axis=-1)).T
    h2norm = (hist2.T - np.mean(hist2, axis=-1)).T

    if np.count_nonzero(h1norm) == 0 or np.count_nonzero(h2norm) == 0:
        raise ValueError("Both histograms are completely flat")

    num = np.sum(h1norm * h2norm, axis=-1)
    denom = np.sqrt(np.sum(h1norm**2, axis=-1) * np.sum(h2norm**2, axis=-1))
    return num / denom


def bincount_batch(x: np.ndarray, axis: int = -1, minlength: int = 0) -> np.ndarray:
    if x.shape[axis] == 0:
        raise ValueError("Specified axis of x cannot be 0")

    minlength = max(minlength, x.max() + 1)
    return np.apply_along_axis(np.bincount, axis, x, minlength=minlength)


def stochastic(x: np.ndarray) -> np.ndarray:
    """It normalizes the last dimension of an ndarray to sum to 1.
    It can be used to convert (batches of) vectors to stochastic vectors
    or (batches of) matrices to *right* stochastic matrices.
    Right stochastic matrices are also called transitions matrices.
    """

    n = np.linalg.norm(x, 1, axis=-1, keepdims=True)
    # n = np.sum(x, axis=-1, keepdims=True) # todo: same result (except dtype), which is faster?

    with np.errstate(invalid="raise"):  # see: `normalized`
        return x / n


def sequence_mask(lengths: np.ndarray, maxlen: Optional[int] = None, dtype: Any = None) -> np.ndarray:
    """cf. tf.sequence_mask
    lengths: [N]
    """

    if not lengths:
        return np.array([], dtype=dtype or np.bool_)

    if maxlen is None:
        maxlen = max(lengths)
    row_vector = np.arange(maxlen)
    matrix = np.expand_dims(lengths, -1)

    ret = row_vector < matrix
    if dtype is not None:
        ret = ret.astype(dtype)

    return ret


# @opjit() # doesn't work atm because numba doesn't support `None` indexing
def _viterbi_dense_masked(
    p_emit: np.ndarray, p_trans: np.ndarray, p_trans0: np.ndarray, mask: np.ndarray
) -> np.ndarray:
    batch_size, T, N = p_emit.shape

    N1, N2 = p_trans.shape
    assert N == N1 == N2

    assert (batch_size, T) == mask.shape
    assert (N,) == p_trans0.shape

    trellis = np.zeros((T, batch_size, N), dtype=p_emit.dtype)
    states = np.zeros((T, batch_size, N), dtype=np.intp)

    # even if sequences of length 0 should be allowed, this does not have to be masked, as there cannot be a wrong result
    trellis[0] = p_trans0 + p_emit[:, 0, :]  # broadcast p_trans0: [N] -> batch_size*[[N]]

    for t in range(1, T):
        masked_p_trans = (
            mask[:, t, None, None] * p_trans[None, :, :]
        )  # [batch_size, 1, 1] * [1, N, N] == [batch_size, N, N]
        weighted_scores = (
            trellis[t - 1, :, :, None] + masked_p_trans
        )  # [batch_size, N, N] # scores and p_trans broadcasted
        max_scores = np.amax(weighted_scores, axis=1)  # [batch_size, N]
        trellis[t] = max_scores + p_emit[:, t, :]  # [batch_size, N] remember highest score of each path
        states[t] = np.argmax(
            weighted_scores, axis=1
        )  # [batch_size, N] remember index of best path, should be repeated for padding vals

    tokens = np.zeros((T, batch_size), dtype=np.intp)
    tokens[T - 1] = np.argmax(trellis[T - 1], axis=1)  # [batch_size]

    for t in range(T - 1, 0, -1):
        tokens[t - 1] = states[t, np.arange(batch_size), tokens[t]]  # [batch_size]

    return tokens.T  # [batch_size, T]


def viterbi_dense(
    p_emit: np.ndarray, p_trans: np.ndarray, p_trans0: Optional[np.ndarray] = None, mask: Optional[np.ndarray] = None
) -> np.ndarray:
    """Viterbi algorithm for finding the optimal path.
    One square transition matrix can be specified.

    T: max sequence length
    N: vocab size
    batch_size: to vectorize

    p_emit: masked emission scores [batch_size, T, N] float
    p_trans: transition scores [N, N] float
    p_trans0: initial transition scores [N] float
    mask: mask out padding values in transitions [batch_size, T] float
    """

    batch_size, T, N = p_emit.shape

    if mask is None:
        mask = np.ones((batch_size, T), dtype=p_trans.dtype)

    if p_trans0 is None:
        p_trans0 = np.zeros(N, dtype=p_emit.dtype)

    return _viterbi_dense_masked(p_emit, p_trans, p_trans0, mask)


# @opjit() # doesn't work atm because numba cannot store arbitrary elements in lists I think
def viterbi_sparse(p_emit: Sequence[np.ndarray], p_trans: Sequence[np.ndarray]) -> np.ndarray:
    """Viterbi algorithm for finding the optimal path.
    The number of emission probabilities per index can vary
    and a separate matrix can be specified for each transition.

    p_emit: Sequence of [x] matrices, where x can vary for every entry
    p_trans: Sequence of [y, z] matrices, where y and z can vary for every entry

    todo: using `Iterable`s instead of Sequence`s should be possible.
    """

    T = len(p_emit)

    assert T - 1 == len(p_trans)

    trellis = [p_emit[0]]
    states = [None]

    for t in range(1, T):
        weighted_scores = trellis[-1][:, None] + p_trans[t - 1]  # [x, y] # scores and p_trans broadcasted
        max_scores = np.amax(weighted_scores, axis=0)  # [y]
        trellis.append(np.add(max_scores, p_emit[t]))  # [y] remember highest score of each path
        states.append(np.argmax(weighted_scores, axis=0))  # [y] remember index of best path

    assert len(trellis) == T and len(states) == T

    tokens = [None] * T  # [T]
    tokens[-1] = np.argmax(trellis[-1], axis=0)  # []

    for t in range(T - 1, 0, -1):
        tokens[t - 1] = states[t][tokens[t]]  # []

    return tokens


_bit_counts = np.array([int(bin(x).count("1")) for x in range(256)]).astype(np.uint8)


def hamming_dist_packed(a: np.ndarray, b: np.ndarray, axis: Optional[int] = -1) -> np.ndarray:
    return np.sum(_bit_counts[np.bitwise_xor(a, b)], axis=axis)


def get_num_chunks(shape: np.ndarray, chunksize: np.ndarray) -> int:
    return np.prod(np.ceil(shape / chunksize).astype(np.int_))


def broadcast_shapes(*shapes: Tuple[int, ...]) -> Tuple[int, ...]:
    """np.broadcast_shapes requires `numpy==1.20.0`,
    which is not available for `python < 3.7`.
    """

    arrays = [np.empty(shape) for shape in shapes]
    return np.broadcast(*arrays).shape


def hamming_dist_packed_chunked(
    a: np.ndarray, b: np.ndarray, chunksize: Tuple[int, ...], axis: Optional[int] = -1
) -> Iterator[Tuple[Tuple[int, ...], np.ndarray]]:
    outshape = broadcast_shapes(a.shape, b.shape)
    select_broadcasted_axis = slice(0, 1)

    if len(outshape) - len(chunksize) != 1:
        raise ValueError("Length of `chunksize` must be one less the number of input dimensions")

    if len(outshape) == 2:
        for x in range(0, outshape[0], chunksize[0]):
            if a.shape[0] != outshape[0]:
                aix = select_broadcasted_axis
            else:
                aix = slice(x, x + chunksize[0])

            if b.shape[0] != outshape[0]:
                bix = select_broadcasted_axis
            else:
                bix = slice(x, x + chunksize[0])

            yield (x,), hamming_dist_packed(a[aix, :], b[bix, :], axis=axis)

    elif len(outshape) == 3:
        for x in range(0, outshape[0], chunksize[0]):
            for y in range(0, outshape[1], chunksize[1]):
                if a.shape[0] != outshape[0]:
                    aix = select_broadcasted_axis
                else:
                    aix = slice(x, x + chunksize[0])

                if b.shape[0] != outshape[0]:
                    bix = select_broadcasted_axis
                else:
                    bix = slice(x, x + chunksize[0])

                if a.shape[1] != outshape[1]:
                    aiy = select_broadcasted_axis
                else:
                    aiy = slice(y, y + chunksize[1])

                if b.shape[1] != outshape[1]:
                    biy = select_broadcasted_axis
                else:
                    biy = slice(y, y + chunksize[1])

                yield (x, y), hamming_dist_packed(a[aix, aiy, :], b[bix, biy, :], axis=axis)

    else:
        raise ValueError("Input must either be 2 or 3 dimensional")


def hamming_dist(a: np.ndarray, b: np.ndarray, axis: Optional[int] = -1) -> np.ndarray:
    return np.count_nonzero(a != b, axis=axis)


def center_of_mass_2d(arr: np.ndarray, dtype=np.float32) -> np.ndarray:
    """Batched center of mass calculation of 2d arrays
    `arr`: [..., x, y]
    """

    total = np.sum(arr, axis=(-1, -2))
    grids = np.ogrid[[slice(0, i) for i in arr.shape[-2:]]]
    with np.errstate(invalid="ignore"):
        results = np.array([np.sum(arr * grid.astype(dtype), axis=(-1, -2)) / total for grid in grids], dtype=dtype)
    results = np.moveaxis(results, 0, -1)
    return results


def ceildiv(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    return -(a // -b)


def image_grid(images: np.ndarray, nrow: int = 8, fill_value=(0, 0, 0)) -> np.ndarray:
    """Creates a picture grid (left to right, top to bottom).
    `images`: Input array of shape [n, height, widght, channel]
    `nrow`: Number of images per row
    `fill_value`: Pad empty space with pixels of shape [channel]
    """

    if not images.ndim == 4:
        raise ValueError("Input shape must be [n, height, width, channel]")

    h, w = images.shape[1:3]
    nbatch = images.shape[0]
    channels = images.shape[3]
    rows = ceildiv(nbatch, nrow)
    cols = min(nbatch, nrow)

    if channels != len(fill_value):
        raise ValueError("fill_value must match number of channels")

    out = np.full((h * rows, w * cols, channels), fill_value, dtype=images.dtype)

    for i in range(images.shape[0]):
        row = i // cols
        col = i % cols
        out[h * row : h * (row + 1), w * col : w * (col + 1), :] = images[i]

    return out


def limit_balanced(arr: np.ndarray, total: int) -> np.ndarray:
    """Use a list of counts `arr`, return a new list where the total count is `total` (if possible),
    in such a way that the selection is as balanced as possible.
    """

    assert len(arr.shape) == 1

    indices = np.argsort(arr)
    values = arr[indices]
    remaining_total = total
    remaining_cats = len(arr)
    balanced: list[int] = []
    for value in values:
        n_chose = remaining_total // remaining_cats
        if value >= n_chose:
            balanced.append(n_chose)
            remaining_total -= n_chose
        else:
            balanced.append(value)
            remaining_total -= value
        remaining_cats -= 1

    out = np.array(balanced)
    inverse_indices = np.argsort(indices)
    restored_out = out[inverse_indices]

    assert arr.shape == out.shape
    assert np.all(restored_out <= arr)
    assert np.sum(restored_out) == min(total, np.sum(arr))
    return restored_out


def indind(absolute: Union[slice, np.ndarray], relative: np.ndarray) -> np.ndarray:
    if isinstance(absolute, slice):
        assert absolute == slice(None)
        return relative
    else:
        return absolute[relative]


def _arg_sample_balanced(
    arr: np.ndarray,
    indices: Union[slice, np.ndarray],
    total: int,
    dim: int,
    rng: np.random.Generator,
) -> np.ndarray:
    subarray = arr[indices, dim]
    values, counts = np.unique(subarray, return_counts=True)
    n_chose_arr = limit_balanced(counts, total)

    out = []

    if dim == arr.shape[1] - 1:
        # last column, actually select values
        for v, subtotal in zip(values, n_chose_arr):
            (ind,) = np.nonzero(subarray == v)
            if len(ind) != subtotal:
                ind = rng.choice(ind, subtotal, replace=False)
            out.append(indind(indices, ind))
    else:
        # recurse into other columns
        for v, subtotal in zip(values, n_chose_arr):
            (ind,) = np.nonzero(subarray == v)
            out.append(_arg_sample_balanced(arr, indind(indices, ind), subtotal, dim + 1, rng))

    return np.concatenate(out)


def arg_sample_balanced(arr: np.ndarray, total: int, seed: int = 0) -> np.ndarray:
    """Select a `total` number of samples from `arr` and return the indices.
    `arr` needs to be a 2-dimensional array where the first dimension are the samples
    and the second one categorical columns. These categories are used to sample in a way,
    so that the result is a balanced as possible, with using additional columns with decreasing priority.
    The output are the indices used to select the samples. They are in no specific order.
    To select without reordering, sort the output.
    """

    if len(arr.shape) != 2 or arr.shape[0] < 1 or arr.shape[1] < 1:
        raise ValueError("array must be 2-dimensional with at least one element in each dimension")

    dim = 0
    rng = np.random.default_rng(seed=seed)

    out = _arg_sample_balanced(arr, slice(None), total, dim, rng)
    assert len(out) == min(total, len(arr))
    assert len(np.unique(out)) == len(out)
    return out
